@@ -1,6 +1,7 @@
 import json
 import ctypes as ct
-from coreir.type import COREType_p, Type, Params, COREValue_p, Values, Record
+from coreir.base import _cache
+from coreir.type import COREType_p, Type, Params, COREValue_p, Values, Record, Value
 from coreir.generator import Generator
 from coreir.namespace import Namespace, CORENamespace_p
 from coreir.lib import libcoreir_c, load_coreir_lib, libcoreir_sim_c
@@ -43,7 +44,19 @@ class NamedTypesDict:
         )
 
 
-_LIBRARY_CACHE = {}
+_library_cache = {}
+
+def namespace_cache(f):
+    def method(self, name: str):
+        c_addr = ct.addressof(self.context)
+        if name in _library_cache[c_addr]:
+            return _library_cache[c_addr][name]
+        ns = f(self, name)
+        assert isinstance(ns, Namespace)
+        _library_cache[c_addr][name] = ns
+        return ns
+    return method
+
 class Context:
     def __init__(self, ptr=None):
         # FIXME: Rename this to ptr or context_ptr to be consistent with other
@@ -52,11 +65,19 @@ class Context:
         if ptr is None:
             self.external_ptr = False
             ptr = libcoreir_c.CORENewContext()
-            _LIBRARY_CACHE.setdefault(ct.addressof(ptr), {})
+            _library_cache.setdefault(ct.addressof(ptr), {})
         self.context = ptr
-        self.global_namespace = Namespace(libcoreir_c.COREGetGlobal(self.context),self)
+        #self.global_namespace = Namespace(libcoreir_c.COREGetGlobal(self.context),self)
         self.named_types = NamedTypesDict(self)
 
+    def __str__(self):
+        return f"Context<{self.external_ptr}, {self.context}>"
+
+    @property
+    def global_namespace(self):
+        return self.get_lib("global")
+
+
     @property
     def G(self):
         raise Exception("Context.G has been removed, use Context.global_namespace instead")
@@ -116,7 +137,9 @@ class Context:
     def new_values(self,fields={}):
         args = []
         for v in fields.values():
-            if type(v) is int:
+            if isinstance(v, Value):
+                args.append(v.ptr)
+            elif type(v) is int:
                 args.append(libcoreir_c.COREValueInt(self.context, ct.c_int(v)))
             elif type(v) is str:
                 args.append(libcoreir_c.COREValueString(self.context,
@@ -146,6 +169,7 @@ class Context:
         return Values(gen_args,self)
 
     def load_from_file(self, file_name):
+
         err = ct.c_bool(False)
         m = libcoreir_c.CORELoadModule(
                 self.context, ct.c_char_p(str.encode(file_name)),ct.byref(err))
@@ -163,35 +187,39 @@ class Context:
         if err.value is not False:
             raise Exception("Error saving context")
 
+    @namespace_cache
     def load_library(self, name):
-        c_addr = ct.addressof(self.context)
-        if name in _LIBRARY_CACHE[c_addr]:
-            return _LIBRARY_CACHE[c_addr][name]
         lib = load_coreir_lib(name)
         func = getattr(lib,"CORELoadLibrary_{}".format(name))
         func.argtypes = [COREContext_p]
         func.restype = CORENamespace_p
-        ns = Namespace(func(self.context), self)
-        _LIBRARY_CACHE[c_addr][name] = ns
-        return ns
+        return Namespace(func(self.context), self)
 
     def enable_symbol_table(self):
         libcoreir_sim_c.COREEnSymtable(self.context)
 
+    @namespace_cache
     def get_namespace(self,name):
-      ns = libcoreir_c.COREGetNamespace(self.context,ct.c_char_p(str.encode(name)))
-      return Namespace(ns,self)
+        ns = libcoreir_c.COREGetNamespace(self.context,ct.c_char_p(str.encode(name)))
+        return Namespace(ns, self)
+
+    def has_namespace(self, name):
+        return libcoreir_c.COREHasNamespace(self.context,
+                                            ct.c_char_p(str.encode(name)))
 
     def new_namespace(self,name):
-      ns = libcoreir_c.CORENewNamespace(self.context,ct.c_char_p(str.encode(name)))
-      return Namespace(ns,self)
+        c_addr = ct.addressof(self.context)
+        if name in _library_cache[c_addr]:
+            raise ValueError(f"Namespace {name} already exists!")
+        ns = libcoreir_c.CORENewNamespace(self.context,ct.c_char_p(str.encode(name)))
+        return Namespace(ns, self)
 
-    @lru_cache(maxsize=None)
+    @namespace_cache
     def get_lib(self, lib):
-        if lib in {"coreir", "mantle", "corebit", "memory"}:
+        if lib in {"global", "coreir", "mantle", "corebit", "memory"}:
             return self.get_namespace(lib)
-        elif lib == "global":
-            return self.global_namespace
+        #elif lib == "global":
+        #    return self.global_namespace
         else:
             return self.load_library(lib)
 
@@ -225,8 +253,25 @@ class Context:
                                                 disable_width_cast)
 
     def __del__(self):
-        if not self.external_ptr:
-            libcoreir_c.COREDeleteContext(self.context)
+        if self.context is not None and not self.external_ptr:
+            self.delete()
+
+    def delete(self):
+        """
+        WARNING: Unsafe, will set self.context to None.  The pycoreir internal
+        code does not check for this.  For performance reasons, we do not add a
+        None check for all the API code, instead we assume if this is used that
+        the user is certain that the context object will no longer be used.
+        """
+        if self.context is None:
+            raise Exception("Context already deleted")
+        c_addr = ct.addressof(self.context)
+        if c_addr in _library_cache:
+            del _library_cache[c_addr]
+        if c_addr in _cache:
+            del _cache[c_addr]
+        libcoreir_c.COREDeleteContext(self.context)
+        self.context = None
 
     def Int(self):
         return libcoreir_c.COREContextInt(self.context)
